import healpy as hp
import numpy as np
from astropy import units
from scipy.interpolate import interp1d

from .base_observer import BaseObserver
from .base_skymodel import BaseSkyModel
from .component_data import LFSM_DATA_URL, download_from_url_list

T_CMB = 2.725


def rotate_equatorial_to_galactic(map):
    """
    Given a map in equatorial coordinates, convert it to Galactic coordinates.
    """
    rotCG = hp.rotator.Rotator(coord=("C", "G"))
    nSides = hp.pixelfunc.npix2nside(map.size)
    theta, phi = hp.pixelfunc.pix2ang(nSides, range(map.size))
    theta_new, phi_new = rotCG(theta, phi, inv=True)
    map2 = hp.get_interp_val(map, theta_new, phi_new)
    return map2


class LowFrequencySkyModel(BaseSkyModel):
    """LWA1 Low Frequency Sky Model"""

    def __init__(self, freq_unit="MHz", include_cmb=False):
        """Global sky model (GSM) class for generating sky models.

        Parameters
        ----------
        freq_unit (str): Frequency unit to use, defaults to MHz
        include_cmb (bool):  Choose whether to include the CMB. Defaults to False. A value of
                             T_CMB = 2.725 K is used if True.
        """
        data_unit = "K"
        basemap = "LFSS"

        # download component data as needed using astropy cache
        LFSM_FILEPATH = download_from_url_list(LFSM_DATA_URL)

        super(LowFrequencySkyModel, self).__init__(
            "LFSM", LFSM_FILEPATH, freq_unit, data_unit, basemap
        )

        self.pca_map = self.h5["lfsm_component_maps_3.0deg.dat"][:]
        self.pca_components = self.h5["lfsm_components.dat"][:]
        self.nside = 256

        self.include_cmb = include_cmb

        freqs = self.pca_components[:, 0]
        sigmas = self.pca_components[:, 1]
        comps = self.pca_components[:, 2:]

        self.scaleFunc = interp1d(np.log(freqs), np.log(sigmas), kind="slinear")

        self.compFuncs = []
        for i in range(comps.shape[1]):
            self.compFuncs.append(interp1d(np.log(freqs), comps[:, i], kind="cubic"))

    def generate(self, freqs):
        """Generate a global sky model at a given frequency or frequencies

        Parameters
        ----------
        freqs: float or np.array
            Frequency for which to return GSM model

        Returns
        -------
        gsm: np.array
            Global sky model in healpix format, with NSIDE=256. Output map
            is in galactic coordinates, and in antenna temperature units (K).

        """
        # convert frequency values into Hz
        freqs = np.array(freqs) * units.Unit(self.freq_unit)
        freqs_mhz = freqs.to("MHz").value

        if isinstance(freqs_mhz, float):
            freqs_mhz = np.array([freqs_mhz])

        try:
            assert np.min(freqs_mhz) >= 10
            assert np.max(freqs_mhz) <= 408
        except AssertionError:
            raise RuntimeError("Frequency values lie outside 10 MHz < f < 408 MHz")

        map_out = 0.0
        if isinstance(freqs, np.ndarray):
            if freqs.ndim > 0:
                map_out = np.zeros(shape=(freqs.shape[0], hp.nside2npix(self.nside)))
            else:
                map_out = np.zeros(shape=(1, hp.nside2npix(self.nside)))
        else:
            map_out = np.zeros(shape=(1, hp.nside2npix(self.nside)))

        for ff in range(map_out.shape[0]):
            for i, compFunc in enumerate(self.compFuncs):
                map_out[ff] += compFunc(np.log(freqs_mhz[ff])) * self.pca_map[:, i]
            map_out[ff] *= np.exp(self.scaleFunc(np.log(freqs_mhz[ff])))

            map_out[ff] = rotate_equatorial_to_galactic(map_out[ff])

        map_out = map_out.squeeze()

        if self.include_cmb == False:
            map_out -= T_CMB

        self.generated_map_data = map_out
        self.generated_map_freqs = freqs
        return map_out


class LFSMObserver(BaseObserver):
    def __init__(self):
        """Initialize the Observer object.

        Calls ephem.Observer.__init__ function and adds on gsm
        """
        super(LFSMObserver, self).__init__(gsm=LowFrequencySkyModel)



import healpy as hp
import numpy as np
from astropy.coordinates import SkyCoord


def hpix2sky(nside: int, pix_ids: np.ndarray) -> SkyCoord:
    """Convert a healpix pixel_id into a SkyCoord

    Args:
        nside (int): Healpix NSIDE parameter
        pix_ids (np.array): Array of pixel IDs

    Returns:
        sc (SkyCoord): Corresponding SkyCoordinates
    """
    gl, gb = hp.pix2ang(nside, pix_ids, lonlat=True)
    sc = SkyCoord(gl, gb, frame="galactic", unit=("deg", "deg"))
    return sc


def sky2hpix(nside: int, sc: SkyCoord) -> np.ndarray:
    """Convert a SkyCoord into a healpix pixel_id

    Args:
        nside (int): Healpix NSIDE parameter
        sc (SkyCoord): Astropy sky coordinates array

    Returns:
        pix (np.array): Array of healpix pixel IDs
    """
    gl, gb = sc.galactic.l.to("deg").value, sc.galactic.b.to("deg").value
    pix = hp.ang2pix(nside, gl, gb, lonlat=True)
    return pix



"""
test_gsm.py
===========

Tests for GSM module.
"""

import os
import time

import h5py
import healpy as hp
import numpy as np
import pytest
from astropy.coordinates import SkyCoord

from pygdsm import GlobalSkyModel
from pygdsm.component_data import GSM2008_TEST_DATA_URL, download_from_url_list


# Download test data as needed (stored in astropy cache)
GSM2008_TEST_DATA = download_from_url_list(GSM2008_TEST_DATA_URL)

def test_gsm_generate():
    """Test maps generate successfully, and that output shapes are correct."""

    freq = 40
    gsm = GlobalSkyModel()
    map = gsm.generate(freq)
    assert map.shape == (3145728,)

    freqs = [40, 80, 120, 160]
    gsm = GlobalSkyModel(interpolation="cubic")
    map = gsm.generate(freqs)
    assert map.shape == (4, 3145728)

    freqs = np.linspace(1, 90, 10)
    freq_unit = "GHz"
    for map_id in ["5deg", "haslam", "wmap"]:
        gsm = GlobalSkyModel(basemap=map_id, interpolation="pchip", freq_unit=freq_unit)
        map = gsm.generate(freqs)
        assert map.shape == (10, 3145728)


def test_compare_to_gsm():
    """Compare output of python version to fortran version.

    Compares against output of original GSM. Note that the interpolation
    functions used in this and in the original differ. So, the outputs
    differ too, by as much as 5%. This just goes to show that there's only
    so much accuracy that one can get via interpolation.

    If the frequency matches a map used to generate the GSM, then the
    fortran and python should match up pretty much exactly as no interpolation
    is required. In between maps is where the differences will become more apparent.

    Note: The gsm.f that is supplied only ever uses the haslam map. It also
          has no option to change the interpolation method.

    Note: Because each output
    """
    gsm = GlobalSkyModel(basemap="haslam", interpolation="pchip")
    gsm_fortran = h5py.File(GSM2008_TEST_DATA, "r")
    for freq in (10, 100, 408, 1000, 1420, 2326, 10000, 23000, 40000, 90000, 94000):
        print("\nComparing at %i MHz..." % freq)
        a = gsm_fortran["map_%imhz.out" % freq][:]
        b = gsm.generate(freq)
        frac_avg = np.average(np.abs(1 - a / b))
        if frac_avg > 0.01:
            print("FORTRAN: ", a)
            print("PYTHON:  ", b)
            print(f"FRAC AVG: {frac_avg:.6f}")
        assert frac_avg < 0.035


def test_set_methods():
    gsm = GlobalSkyModel()
    gsm.generate(40)
    # gsm.view()

    gsm.set_basemap("haslam")
    gsm.view()

    gsm.set_basemap("5deg")
    gsm.view()

    gsm.set_basemap("wmap")
    gsm.view()

    gsm.set_freq_unit("GHz")
    gsm.view()


def test_speed():
    gsm = GlobalSkyModel(basemap="haslam")

    t1 = time.time()
    for ff in np.linspace(10, 10000, 100):
        gsm.generate(ff)
    t2 = time.time()
    print("Time taken: %2.2fs" % (t2 - t1))


def test_write_fits():
    gsm = GlobalSkyModel()
    gsm.generate(1000)
    gsm.write_fits("test_write_fits.fits")

    d_fits = hp.read_map("test_write_fits.fits")
    d_gsm = gsm.generated_map_data

    assert d_fits.shape == d_gsm.shape
    assert np.allclose(d_fits, d_gsm)

    os.remove("test_write_fits.fits")


def test_cmb_removal():
    g = GlobalSkyModel(freq_unit="MHz", include_cmb=False)
    sky_no_cmb = g.generate(400)
    g = GlobalSkyModel(freq_unit="MHz", include_cmb=True)
    sky_with_cmb = g.generate(400)

    T_cmb = (sky_with_cmb - sky_no_cmb).mean()
    print(T_cmb)
    assert np.isclose(T_cmb, 2.725)


def test_get_sky_temperature():
    gc = SkyCoord(0, 0, unit="deg", frame="galactic")
    freqs = (50, 100, 150)
    g = GlobalSkyModel()
    T = g.get_sky_temperature(gc, freqs)

    T_gold = np.array([258368.7463149, 50466.45058671, 18968.12555978])
    assert np.allclose(T, T_gold)


def test_stupid_values():
    with pytest.raises(RuntimeError):
        g = GlobalSkyModel(basemap="haslamalan")
    with pytest.raises(RuntimeError):
        g = GlobalSkyModel(interpolation="linear")
    with pytest.raises(RuntimeError):
        g = GlobalSkyModel()
        g.generate(9999999999)


def test_set_interpolation_method():
    g = GlobalSkyModel(interpolation="pchip")
    assert g.interpolation_method == "pchip"
    g.set_interpolation_method("cubic")
    assert g.interpolation_method == "cubic"


if __name__ == "__main__":
    test_stupid_values()
    test_set_interpolation_method()
    test_gsm_generate()
    test_compare_to_gsm()
    test_speed()
    test_write_fits()
    test_set_methods()
    test_cmb_removal()
    test_get_sky_temperature()



import os
from datetime import datetime

import healpy as hp
import numpy as np
import pylab as plt
from astropy.time import Time
import pytest

from pygdsm import GSMObserver, GSMObserver16, LFSMObserver


def test_gsm_observer(show=False):
    """Test GSMObserver() is working"""
    (latitude, longitude, elevation) = ("37.2", "-118.2", 1222)
    ov = GSMObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)
    ov.generate(50)
    ov.view(logged=True)
    ov.view_observed_gsm(logged=True)
    if show:
        plt.show()

    (latitude, longitude, elevation) = ("37.2", "-118.2", 1222)
    ov = GSMObserver16()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)
    ov.generate(50)
    ov.view(logged=True)
    ov.view_observed_gsm(logged=True)
    if show:
        plt.show()

    (latitude, longitude, elevation) = ("37.2", "-118.2", 1222)
    ov = LFSMObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)
    ov.generate(50)
    ov.view(logged=True)
    ov.view_observed_gsm(logged=True)

    with pytest.raises(ValueError) as e:
        ov.generate(horizon_elevation=-1e-3)

    if show:
        plt.show()


def test_observed_mollview():
    """Generate animated maps of observing coverage over 24 hours"""

    (latitude, longitude, elevation) = ("37.2", "-118.2", 1222)
    ov = GSMObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation

    obs = []
    if not os.path.exists('generated_sky'):
        os.mkdir('generated_sky')
    freq = 50
    horizon_elevation = '85.0'
    for ii in range(0, 24, 4):
        ov.date = datetime(2000, 1, 1, ii, 0)
        ov.generate(freq)
        sky = ov.view_observed_gsm(logged=True, show=False, min=9, max=20)
        plt.savefig("generated_sky/galactic-%02d.png" % ii)
        plt.close()

        hp.mollview(sky, coord=["G", "E"], min=9, max=20)
        plt.savefig("generated_sky/ecliptic-%02d.png" % ii)
        plt.close()

        hp.mollview(sky, coord=["G", "C"], min=9, max=20)
        plt.savefig("generated_sky/equatorial-%02d.png" % ii)
        plt.close()

        ov.view(logged=True, show=False, min=9, max=20)
        plt.savefig("generated_sky/ortho-%02d.png" % ii)
        plt.close()

        ov.generate(freq=freq, horizon_elevation=horizon_elevation)
        ov.view(logged=True, show=False, min=9, max=20)
        plt.savefig('generated_sky/ortho_85deg_horizon-%02d.png' % ii)
        plt.close()

        ov.generate(freq=freq, horizon_elevation=horizon_elevation)
        ov.view(logged=True, show=False, min=9, max=20)
        plt.savefig('generated_sky/ortho_85deg_horizon-%02d.png' % ii)
        plt.close()
        print(ii)

    os.system('convert -delay 20 generated_sky/ortho-*.png ortho.gif')
    os.system('convert -delay 20 generated_sky/ortho_85deg_horizon-*.png ortho_85deg_horizon.gif')
    os.system('convert -delay 20 generated_sky/galactic-*.png galactic.gif')
    os.system('convert -delay 20 generated_sky/ecliptic-*.png ecliptic.gif')
    os.system('convert -delay 20 generated_sky/equatorial-*.png equatorial.gif')



def test_generate_with_and_without_args():
    """Test generating without frequency argument"""
    (latitude, longitude, elevation) = ("37.2", "-118.2", 1222)
    ov = GSMObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    ov.generate(50)

    ov.date = datetime(2000, 1, 1, 22, 0)
    ov.generate()

    ov.generate(51)
    ov.generate(52)
    ov.generate()
    now = Time(datetime.now())
    ov.generate(obstime=now)
    ov.generate(obstime=now)
    ov.generate(obstime=now, freq=53)
    ov.generate(obstime=now, freq=52)
    ov.generate(obstime=now, freq=53, horizon_elevation=0.0)
    ov.generate(obstime=now, freq=52, horizon_elevation='0.0')
    ov.generate(obstime=now, freq=53, horizon_elevation=np.deg2rad(85.0))
    ov.generate(obstime=now, freq=52, horizon_elevation='85.0')

if __name__ == "__main__":
    test_gsm_observer(show=True)
    test_observed_mollview()
    test_generate_with_and_without_args()



"""
test_init.py
===========

Tests for GSM init commands
"""

from pygdsm import (
    GlobalSkyModel,
    GlobalSkyModel08,
    GlobalSkyModel16,
    GSMObserver,
    GSMObserver16,
    HaslamObserver,
    HaslamSkyModel,
    LFSMObserver,
    LowFrequencySkyModel,
    init_gsm,
    init_observer,
)


def test_init():
    assert isinstance(init_gsm("gsm"), GlobalSkyModel)
    assert isinstance(init_gsm("gsm08"), GlobalSkyModel)
    assert isinstance(init_gsm("gsm16"), GlobalSkyModel16)
    assert isinstance(init_gsm("lfsm"), LowFrequencySkyModel)
    assert isinstance(init_gsm("haslam"), HaslamSkyModel)

    assert isinstance(init_observer("gsm"), GSMObserver)
    assert isinstance(init_observer("gsm08"), GSMObserver)
    assert isinstance(init_observer("gsm16"), GSMObserver16)
    assert isinstance(init_observer("lfsm"), LFSMObserver)
    assert isinstance(init_observer("haslam"), HaslamObserver)

    gsm16 = init_gsm(
        "gsm16",
        freq_unit="MHz",
        include_cmb=True,
        interpolation="PCHIP",
    )
    assert isinstance(gsm16, GlobalSkyModel16)

if __name__ == "__main__":
    test_init()



import healpy as hp
import numpy as np
from astropy import units
from scipy.interpolate import interp1d, pchip

from .base_observer import BaseObserver
from .base_skymodel import BaseSkyModel
from .component_data import GSM2016_DATA_URL, download_from_url_list

kB = 1.38065e-23
C = 2.99792e8
h = 6.62607e-34
T = 2.725
hoverk = h / kB


def K_CMB2MJysr(K_CMB, nu):  # in Kelvin and Hz
    B_nu = 2 * (h * nu) * (nu / C) ** 2 / (np.exp(hoverk * nu / T) - 1)
    conversion_factor = (B_nu * C / nu / T) ** 2 / 2 * np.exp(hoverk * nu / T) / kB
    return K_CMB * conversion_factor * 1e20  # 1e-26 for Jy and 1e6 for MJy


def K_RJ2MJysr(K_RJ, nu):  # in Kelvin and Hz
    conversion_factor = 2 * (nu / C) ** 2 * kB
    return K_RJ * conversion_factor * 1e20  # 1e-26 for Jy and 1e6 for MJy


def rotate_map(hmap, rot_theta, rot_phi, nest=True):
    nside = hp.npix2nside(len(hmap))

    # Get theta, phi for non-rotated map
    t, p = hp.pix2ang(nside, np.arange(hp.nside2npix(nside)), nest=nest)  # theta, phi

    # Define a rotator
    r = hp.Rotator(deg=False, rot=[rot_phi, rot_theta])

    # Get theta, phi under rotated co-ordinates
    trot, prot = r(t, p)

    # Inerpolate map onto these co-ordinates
    rot_map = hp.get_interp_val(hmap, trot, prot, nest=nest)

    return rot_map


class GlobalSkyModel16(BaseSkyModel):
    """Global sky model (GSM) class for generating sky models."""

    def __init__(
        self,
        freq_unit="MHz",
        data_unit="TCMB",
        resolution="hi",
        theta_rot=0,
        phi_rot=0,
        interpolation="pchip",
        include_cmb=False,
    ):
        """Global sky model (GSM) class for generating sky models.

        Upon initialization, the map PCA data are loaded into memory and interpolation
        functions are pre-computed.

        Parameters
        ----------
        freq_unit: 'Hz', 'MHz', or 'GHz'
            Unit of frequency. Defaults to 'MHz'.
        data_unit: 'MJysr', 'TCMB', 'TRJ'
            Unit of output data. MJy/Steradian, T_CMB in Kelvin, or T_RJ.
        resolution: 'hi' or 'low'
            Resolution of output map. Either 300 arcmin (low) or 24 arcmin (hi).
            For frequencies under 10 GHz, output is 48 arcmin.
        interpolation: 'cubic' or 'pchip'
            Choose whether to use cubic spline interpolation or
            piecewise cubic hermitian interpolating polynomial (PCHIP).
            PCHIP is designed to never locally overshoot data, whereas
            splines are designed to have smooth first and second derivatives.
        include_cmb: bool (default False)
            Choose whether to include the CMB. Defaults to False. A value of
            T_CMB = 2.725 K is used.

        Notes
        -----
        The scipy `interp1d` function does not allow one to explicitly
        set second derivatives to zero at the endpoints, as is done in
        the original GSM. As such, results will differ. Further, we default
        to use PCHIP interpolation.


        """

        # download component data as needed using astropy cache
        GSM2016_FILEPATH = download_from_url_list(GSM2016_DATA_URL)

        if data_unit not in ["MJysr", "TCMB", "TRJ"]:
            raise RuntimeError(
                "UNIT ERROR: %s not supported. Only MJysr, TCMB, TRJ are allowed."
                % data_unit
            )

        if resolution.lower() in ("hi", "high", "h"):
            resolution = "hi"
        elif resolution.lower() in ("low", "lo", "l"):
            resolution = "low"
        else:
            raise RuntimeError(
                "RESOLUTION ERROR: Must be either hi or low, not %s" % resolution
            )

        super(GlobalSkyModel16, self).__init__(
            "GSM2016", GSM2016_FILEPATH, freq_unit, data_unit, basemap=""
        )

        self.interpolation_method = interpolation
        self.resolution = resolution
        self.include_cmb = include_cmb

        # Map data to load
        labels = ["Synchrotron", "CMB", "HI", "Dust1", "Dust2", "Free-Free"]
        self.n_comp = len(labels)

        if resolution == "hi":
            self.nside = 1024
            self.map_ni = np.array([self.h5["highres_%s_map" % lb][:] for lb in labels])
        else:
            self.nside = 64
            self.map_ni = np.array(self.h5["lowres_maps"])

        self.spec_nf = self.h5["spectra"][:]

        if theta_rot or phi_rot:
            for i, map in enumerate(self.map_ni):
                self.map_ni[i] = rotate_map(map, theta_rot, phi_rot, nest=True)

    def generate(self, freqs):
        """Generate a global sky model at a given frequency or frequencies

        Parameters
        ----------
        freqs: float or np.array
            Frequency for which to return GSM model

        Returns
        -------
        gsm: np.array
            Global sky model in healpix format, with NSIDE=1024. Output map
            is in galactic coordinates, ring format.

        """

        # convert frequency values into Hz
        freqs = np.array(freqs) * units.Unit(self.freq_unit)
        freqs_ghz = freqs.to("GHz").value

        if isinstance(freqs_ghz, float):
            freqs_ghz = np.array([freqs_ghz])

        try:
            assert np.min(freqs_ghz) >= 0.01
            assert np.max(freqs_ghz) <= 5000
        except AssertionError:
            raise RuntimeError("Frequency values lie outside 10 MHz < f < 5 THz: %s")

        map_ni = self.map_ni
        # if self.resolution == 'hi':
        #     map_ni = self.map_ni_hr
        # else:
        #     map_ni = self.map_ni_lr

        spec_nf = self.spec_nf
        nfreq = spec_nf.shape[1]

        # Now borrow code from the orignal GSM2008 model to do a sensible interpolation

        pca_freqs_ghz = spec_nf[0]
        pca_scaling = spec_nf[1]
        pca_comps = spec_nf[2:]
        # Interpolate to the desired frequency values
        ln_pca_freqs = np.log(pca_freqs_ghz)
        if self.interpolation_method == "cubic":
            spl_scaling = interp1d(ln_pca_freqs, np.log(pca_scaling), kind="cubic")
            spl1 = interp1d(ln_pca_freqs, pca_comps[0], kind="cubic")
            spl2 = interp1d(ln_pca_freqs, pca_comps[1], kind="cubic")
            spl3 = interp1d(ln_pca_freqs, pca_comps[2], kind="cubic")
            spl4 = interp1d(ln_pca_freqs, pca_comps[3], kind="cubic")
            spl5 = interp1d(ln_pca_freqs, pca_comps[4], kind="cubic")
            spl6 = interp1d(ln_pca_freqs, pca_comps[5], kind="cubic")

        else:
            spl_scaling = pchip(ln_pca_freqs, np.log(pca_scaling))
            spl1 = pchip(ln_pca_freqs, pca_comps[0])
            spl2 = pchip(ln_pca_freqs, pca_comps[1])
            spl3 = pchip(ln_pca_freqs, pca_comps[2])
            spl4 = pchip(ln_pca_freqs, pca_comps[3])
            spl5 = pchip(ln_pca_freqs, pca_comps[4])
            spl6 = pchip(ln_pca_freqs, pca_comps[5])

        self.interp_comps = (spl_scaling, spl1, spl2, spl3, spl4, spl5, spl6)

        ln_freqs = np.log(freqs_ghz)
        comps = np.row_stack((
            spl1(ln_freqs),
            spl2(ln_freqs),
            spl3(ln_freqs),
            spl4(ln_freqs),
            spl5(ln_freqs),
            spl6(ln_freqs),
        ))
        scaling = np.exp(spl_scaling(ln_freqs))

        # Finally, compute the dot product via einsum (awesome function)
        # c=comp, f=freq, p=pixel. We want to dot product over c for each freq
        # print comps.shape, self.pca_map_data.shape, scaling.shape

        output = np.single(np.einsum("cf,pc,f->fp", comps, map_ni.T, scaling))

        for ifreq, freq in enumerate(freqs_ghz):
            output[ifreq] = hp.pixelfunc.reorder(output[ifreq], n2r=True)

            # DCP 2024.03.29 - Add CMB if requested
            if self.include_cmb:
                output[ifreq] += K_CMB2MJysr(T, 1e9 * freq)

            if self.data_unit == "TCMB":
                conversion = 1.0 / K_CMB2MJysr(1.0, 1e9 * freq)
            elif self.data_unit == "TRJ":
                conversion = 1.0 / K_RJ2MJysr(1.0, 1e9 * freq)
            else:
                conversion = 1.0
            output[ifreq] *= conversion

        if len(output) == 1:
            output = output[0]
        # else:
        #    map_data = np.row_stack(output)

        self.generated_map_freqs = freqs
        self.generated_map_data = output

        return output


class GSMObserver16(BaseObserver):
    def __init__(self):
        """Initialize the Observer object.

        Calls ephem.Observer.__init__ function and adds on gsm
        """
        super(GSMObserver16, self).__init__(gsm=GlobalSkyModel16)



import ephem
import healpy as hp
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.time import Time

from pygdsm.plot_utils import show_plt
from pygdsm.utils import hpix2sky, sky2hpix


class BaseObserver(ephem.Observer):
    """Observer of the Global Sky Model.

    Generates the Observed sky, for a given point on Earth.
    Applies the necessary rotations and coordinate transformations
    so that the observed 'sky' can be returned, instead of the
    full galaxy-centered GSM.

    This class is based on pyephem's Observer(). The GSM bit can be thought of
    as an 'add on' to ephem.Observer, adding the methods generate() and  view(),
    which allows all-sky images for a given point on earth to be produced.
    """

    def __init__(self, gsm):
        """Initialize the Observer object.

        Calls ephem.Observer.__init__ function and adds on gsm
        """
        super(BaseObserver, self).__init__()
        self.observed_sky = None
        self.gsm = gsm()
        self._setup()

    def _setup(self):
        self._freq = 100
        self._time = Time(self.date.datetime())
        # Generate mapping from pix <-> angles
        self.gsm.generate(self._freq)
        self._n_pix = hp.get_map_size(self.gsm.generated_map_data)
        self._n_side = hp.npix2nside(self._n_pix)
        self._theta, self._phi = hp.pix2ang(self._n_side, np.arange(self._n_pix))

        self._pix0 = None
        self._mask = None
        self._horizon_elevation = 0.0
        self._observed_ra = None
        self._observed_dec = None

    def generate(self, freq=None, obstime=None, horizon_elevation=None):
        """ Generate the observed sky for the observer, based on the GSM.

        Parameters
        ----------
        freq: float
            Frequency of map to generate, in units of MHz (default).
        obstime: astropy.time.Time
            Time of observation to generate
        horizon_elevation: float
            Elevation of the artificial horizon (default 0.0)

        Returns
        -------
        observed_sky: np.array
            Numpy array representing the healpix image, centered on zenith,
            with below the horizon masked.
        """
        # Check to see if frequency has changed.
        if freq is not None:
            if not np.isclose(freq, self._freq):
                self.gsm.generate(freq)
                self._freq = freq

        sky = self.gsm.generated_map_data

        # Check if time has changed -- astropy allows None == Time() comparison
        if obstime == self._time or obstime is None:
            time_has_changed = False
        else:
            time_has_changed = True
            self._time = Time(obstime)  # This will catch datetimes, but Time() object should be passed
            self.date = obstime.to_datetime()

        # Match pyephem convertion -- string is degrees, int/float is rad
        horizon_elevation = ephem.degrees(horizon_elevation or 0.0)
        if self._horizon_elevation == horizon_elevation:
            horizon_has_changed = False
        else:
            self._horizon_elevation = horizon_elevation
            horizon_has_changed = True

        # Checking this separately encase someone tries to be smart and sets both the attribute and kwarg
        if self._horizon_elevation < 0:
            raise ValueError(f"Horizon elevation must be greater or equal to 0 degrees (currently {np.rad2deg(horizon_elevation)}).")

        # Match pyephem convertion -- string is degrees, int/float is rad
        horizon_elevation = ephem.degrees(horizon_elevation or 0.0)
        if self._horizon_elevation == horizon_elevation:
            horizon_has_changed = False
        else:
            self._horizon_elevation = horizon_elevation
            horizon_has_changed = True

        # Checking this separately encase someone tries to be smart and sets both the attribute and kwarg
        if self._horizon_elevation < 0:
            raise ValueError(f"Horizon elevation must be greater or equal to 0 degrees (currently {np.rad2deg(horizon_elevation)}).")

        # Rotation is quite slow, only recompute if time or frequency has changed, or it has never been run
        if time_has_changed or self.observed_sky is None or horizon_has_changed:
            # Get RA and DEC of zenith
            ra_zen, dec_zen = self.radec_of(0, np.pi / 2)
            sc_zen = SkyCoord(ra_zen, dec_zen, unit=("rad", "rad"))
            pix_zen = sky2hpix(self._n_side, sc_zen)
            vec_zen = hp.pix2vec(self._n_side, pix_zen)

            # Convert to degrees
            ra_zen *= 180 / np.pi
            dec_zen *= 180 / np.pi

            # Generate below-horizon mask using query_disc
            mask = np.ones(shape=self._n_pix, dtype='bool')
            pix_visible = hp.query_disc(self._n_side, vec=vec_zen, radius=np.pi/2 - self._horizon_elevation)
            mask[pix_visible] = 0
            self._mask = mask

            # Transform from Galactic coordinates to Equatorial
            rot = hp.Rotator(coord=["G", "C"])
            eq_theta, eq_phi = rot(self._theta, self._phi)

            # Convert from Equatorial colatitude and longitude to normal RA and DEC
            dec = 90.0 - np.abs(eq_theta * (180 / np.pi))
            ra = ((eq_phi + 2 * np.pi) % (2 * np.pi)) * (180 / np.pi)

            # Apply rotation to convert from Galactic to Equatorial and center on zenith
            hrot = hp.Rotator(rot=[ra_zen, dec_zen], coord=["G", "C"], inv=True)
            g0, g1 = hrot(self._theta, self._phi)
            pix0 = hp.ang2pix(self._n_side, g0, g1)
            self._pix0 = pix0

            dec_rotated = dec[self._pix0]
            ra_rotated = ra[self._pix0]
            self._observed_ra = ra_rotated
            self._observed_dec = dec_rotated

        sky_rotated = sky[self._pix0]
        mask_rotated = self._mask[self._pix0]

        self.observed_sky = hp.ma(sky_rotated)
        self.observed_sky.mask = mask_rotated

        return self.observed_sky

    def view(self, logged=False, show=False, **kwargs):
        """View the local sky, in orthographic projection.

        Parameters
        ----------
        logged: bool
            Default False, return the log2 image
        """
        sky = self.observed_sky
        if logged:
            sky = np.log2(sky)

        hp.orthview(sky, half_sky=True, **kwargs)

        if show:
            show_plt()
        return sky

    @property
    def observed_gsm(self):
        """Return the GSM (Mollweide), with below-horizon area masked."""
        sky = self.observed_sky

        # Get RA and DEC of zenith
        ra_rad, dec_rad = self.radec_of(0, np.pi / 2)
        ra_deg = ra_rad / np.pi * 180
        dec_deg = dec_rad / np.pi * 180

        # Apply rotation
        derotate = hp.Rotator(rot=[ra_deg, dec_deg])
        g0, g1 = derotate(self._theta, self._phi)
        pix0 = hp.ang2pix(self._n_side, g0, g1)
        sky = sky[pix0]

        coordrotate = hp.Rotator(coord=["C", "G"], inv=True)
        g0, g1 = coordrotate(self._theta, self._phi)
        pix0 = hp.ang2pix(self._n_side, g0, g1)
        sky = sky[pix0]
        return sky

    def view_observed_gsm(self, logged=False, show=False, **kwargs):
        """View the GSM (Mollweide), with below-horizon area masked.

        Args:
            logged (bool): Apply log2 to data (default False)
            show (bool): Call plt.show() (default False)

        Returns:
            sky (np.array): Healpix map of observed GSM.
        """
        sky = self.observed_gsm

        if logged:
            sky = np.log2(sky)

        hp.mollview(sky, coord="G", **kwargs)
        if show:
            show_plt()
        return sky



import healpy as hp
import numpy as np

from pygdsm.utils import hpix2sky, sky2hpix


def test_pix2sky():
    """Small test routine for converting healpix pixel_id to and from SkyCoords"""
    NSIDE = 32
    pix = np.arange(hp.nside2npix(NSIDE))
    sc = hpix2sky(NSIDE, pix)
    pix_roundtrip = sky2hpix(NSIDE, sc)
    assert np.allclose(pix, pix_roundtrip)


if __name__ == "__main__":
    test_pix2sky()



"""
gsm.py
======

Python interface for the Global Sky Model (GSM) or Oliveira-Costa et. al.
This is a python-based equivalent to the Fortran `gsm.f` that comes with the
original data. Instead of the original ASCII DAT files that contain the PCA
data, data are stored in HDF5, which is more efficient.

References
----------
A. de Oliveira-Costa, M. Tegmark, B.M. Gaensler, J. Jonas, T.L. Landecker and P. Reich
A model of diffuse Galactic radio emission from 10 MHz to 100 GHz
Mon. Not. R. Astron. Soc. 388, 247-260 (2008)
doi:10.111/j.1365-2966.2008.13376.x

PCA data from: space.mit.edu/home/angelica/gsm

"""

import numpy as np
from astropy import units
from scipy.interpolate import interp1d, pchip

from .base_observer import BaseObserver
from .base_skymodel import BaseSkyModel
from .component_data import GSM_DATA_URL, download_from_url_list
from .plot_utils import show_plt

T_CMB = 2.725

class GlobalSkyModel(BaseSkyModel):
    """Global sky model (GSM) class for generating sky models."""

    def __init__(
        self,
        freq_unit="MHz",
        basemap="haslam",
        interpolation="pchip",
        include_cmb=False,
    ):
        """Global sky model (GSM) class for generating sky models.

        Upon initialization, the map PCA data are loaded into memory and interpolation
        functions are pre-computed.

        Parameters
        ----------
        freq_unit: 'Hz', 'MHz', or 'GHz'
            Unit of frequency. Defaults to 'MHz'.
        basemap: 'haslam', 'wmap' or '5deg'
            GSM version to generate. The 5deg map has 5.1 degree resolution.
            This is a synthesized map made of all the maps considered in
            the de Oliveira-Costa et. al. paper
            At frequencies below 1GHz, haslam is preferred; you get higher
            resolution (1 degree) by locking to the Haslam 408 MHz map.
            At CMB frequencies, it is best to lock to the WMAP 23 GHz
            map, which is presented denoised with 2 degree resolution.
        interpolation: 'cubic' or 'pchip'
            Choose whether to use cubic spline interpolation or
            piecewise cubic hermitian interpolating polynomial (PCHIP).
            PCHIP is designed to never locally overshoot data, whereas
            splines are designed to have smooth first and second derivatives.
        include_cmb: bool (default False)
            Choose whether to include the CMB. Defaults to False. A value of
            T_CMB = 2.725 K is used.

        Notes
        -----
        The scipy `interp1d` function does not allow one to explicitly
        set second derivatives to zero at the endpoints, as is done in
        the original GSM. As such, results will differ. Further, we default
        to use PCHIP interpolation.

        """

        # download component data as needed using astropy cache
        GSM_FILEPATH = download_from_url_list(GSM_DATA_URL)

        try:
            assert basemap in {"5deg", "wmap", "haslam"}
        except AssertionError:
            raise RuntimeError(
                "GSM basemap unknown: %s. Choose '5deg', 'haslam' or 'wmap'" % basemap
            )

        try:
            assert interpolation in {"cubic", "pchip"}
        except AssertionError:
            raise RuntimeError("Interpolation must be set to either 'cubic' or 'pchip'")

        data_unit = "K"
        super(GlobalSkyModel, self).__init__(
            "GSM2008", GSM_FILEPATH, freq_unit, data_unit, basemap
        )

        self.interpolation_method = interpolation
        self.include_cmb = include_cmb

        self.pca_map_data = None
        self.interp_comps = None
        self.update_interpolants()

        self.generated_map_data = None
        self.generated_map_freqs = None
        self.nside = 512

    def update_interpolants(self):
        # Choose the PCA map to load from the HDF5 file
        pca_map_dict = {
            "5deg": "component_maps_5deg",
            "haslam": "component_maps_408locked",
            "wmap": "component_maps_23klocked",
        }
        pca_map_key = pca_map_dict[self.basemap]
        self.pca_map_data = self.h5[pca_map_key][:]

        # Now, load the PCA eigenvalues
        pca_table = self.h5["components"][:]
        pca_freqs_mhz = pca_table[:, 0]
        pca_scaling = pca_table[:, 1]
        pca_comps = pca_table[:, 2:].T

        # Interpolate to the desired frequency values
        ln_pca_freqs = np.log(pca_freqs_mhz)

        if self.interpolation_method == "cubic":
            spl_scaling = interp1d(ln_pca_freqs, np.log(pca_scaling), kind="cubic")
            spl1 = interp1d(ln_pca_freqs, pca_comps[0], kind="cubic")
            spl2 = interp1d(ln_pca_freqs, pca_comps[1], kind="cubic")
            spl3 = interp1d(ln_pca_freqs, pca_comps[2], kind="cubic")

        else:
            spl_scaling = pchip(ln_pca_freqs, np.log(pca_scaling))
            spl1 = pchip(ln_pca_freqs, pca_comps[0])
            spl2 = pchip(ln_pca_freqs, pca_comps[1])
            spl3 = pchip(ln_pca_freqs, pca_comps[2])
        self.interp_comps = (spl_scaling, spl1, spl2, spl3)

    def generate(self, freqs):
        """Generate a global sky model at a given frequency or frequencies

        Parameters
        ----------
        freqs: float or np.array
            Frequency for which to return GSM model

        Returns
        -------
        gsm: np.array
            Global sky model in healpix format, with NSIDE=512. Output map
            is in galactic coordinates, and in antenna temperature units (K).

        """
        # convert frequency values into Hz
        freqs = np.array(freqs) * units.Unit(self.freq_unit)
        freqs_mhz = freqs.to("MHz").value

        if isinstance(freqs_mhz, float):
            freqs_mhz = np.array([freqs_mhz])

        try:
            assert np.min(freqs_mhz) >= 10
            assert np.max(freqs_mhz) <= 94000
        except AssertionError:
            raise RuntimeError("Frequency values lie outside 10 MHz < f < 94 GHz")

        # Load interpolators and do interpolation
        ln_freqs = np.log(freqs_mhz)
        spl_scaling, spl1, spl2, spl3 = self.interp_comps
        comps = np.row_stack((spl1(ln_freqs), spl2(ln_freqs), spl3(ln_freqs)))
        scaling = np.exp(spl_scaling(ln_freqs))

        # Finally, compute the dot product via einsum (awesome function)
        # c=comp, f=freq, p=pixel. We want to dot product over c for each freq
        # print comps.shape, self.pca_map_data.shape, scaling.shape
        map_out = np.einsum("cf,pc,f->fp", comps, self.pca_map_data, scaling)

        if self.include_cmb:
            map_out += T_CMB

        if map_out.shape[0] == 1:
            map_out = map_out[0]
        self.generated_map_data = map_out
        self.generated_map_freqs = freqs

        return map_out

    def set_basemap(self, new_basemap):
        self.basemap = new_basemap
        self.update_interpolants()
        if self.generated_map_freqs is not None:
            self.generate(self.generated_map_freqs)

    def set_freq_unit(self, new_unit):
        self.freq_unit = new_unit
        self.update_interpolants()
        if self.generated_map_freqs is not None:
            self.generate(self.generated_map_freqs)

    def set_interpolation_method(self, new_method):
        self.interpolation_method = new_method
        self.update_interpolants()
        if self.generated_map_freqs is not None:
            self.generate(self.generated_map_freqs)


class GSMObserver(BaseObserver):
    def __init__(self):
        """Initialize the Observer object.

        Calls ephem.Observer.__init__ function and adds on gsm
        """
        super(GSMObserver, self).__init__(gsm=GlobalSkyModel)



import os

import h5py
import healpy as hp
import numpy as np
from astropy.io import fits

from .plot_utils import show_plt


def is_fits(filepath):
    """
    Check if file is a FITS file
    Returns True of False

    Parameters
    ----------
    filepath: str
        Path to file
    """
    FITS_SIGNATURE = (
        b"\x53\x49\x4d\x50\x4c\x45\x20\x20\x3d\x20\x20\x20\x20\x20"
        b"\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20"
        b"\x20\x54"
    )
    try:
        with open(str(filepath), "rb") as f:
            return f.read(30) == FITS_SIGNATURE
    except FileNotFoundError as e:
        print(e)
        return False


class BaseSkyModel(object):
    """Global sky model (GSM) class for generating sky models."""

    def __init__(self, name, filepath, freq_unit, data_unit, basemap):
        """Initialise basic sky model class

        Parameters
        ----------
        name (str):      Name of GSM
        filepath (str):  Path to HDF5 data / FITS data (healpix) to load
        freq_unit (str): Frequency unit (MHz / GHz / Hz)
        data_unit (str): Unit for pixel scale (e.g. K)
        basemap (str):   Map used as a basis for spatial structure in PCA fit.

        Notes
        -----
        Any GSM needs to supply a generate() function
        """
        self.name = name
        if os.path.exists(filepath):
            if h5py.is_hdf5(filepath):
                self.h5 = h5py.File(filepath, "r")
            elif is_fits(filepath):
                self.fits = fits.open(filepath, "readonly")
            else:
                raise RuntimeError(f"Cannot read HDF5/FITS file {filepath}")
        else:
            raise RuntimeError(f"Cannot find {filepath}")
        self.basemap = basemap
        self.freq_unit = freq_unit
        self.data_unit = data_unit

        self.generated_map_data = None
        self.generated_map_freqs = None

    def generate(self, freqs):
        raise NotImplementedError

    def view(self, idx=0, logged=False, show=False):
        """View generated map using healpy's mollweide projection.

        Parameters
        ----------
        idx: int
            index of map to view. Only required if you generated maps at
            multiple frequencies.
        logged: bool
            Take the log of the data before plotting. Defaults to False.

        """

        if self.generated_map_data is None:
            raise RuntimeError(
                "No GSM map has been generated yet. Run generate() first."
            )

        if self.generated_map_data.ndim == 2:
            gmap = self.generated_map_data[idx]
            freq = self.generated_map_freqs[idx]
        else:
            gmap = self.generated_map_data
            freq = self.generated_map_freqs

        if logged:
            gmap = np.log2(gmap)

        hp.mollview(
            gmap, coord="G", title="%s %s, %s" % (self.name, str(freq), self.basemap)
        )

        if show:
            show_plt()

    def get_sky_temperature(self, coords, freqs=None, include_cmb=True):
        """Get sky temperature at given coordinates.

        Returns sky temperature at a given SkyCoord (e.g. Ra/Dec or galactic l/b).
        Useful for estimating sky contribution to system temperature.

        Parameters
        ----------
        coords (astropy.coordinates.SkyCoord):
            Sky Coordinates to compute temperature for.
        freqs (None, float, or np.array):
            frequencies to evaluate. If not set, will default to those supplied
            when generate() was called.
        include_cmb (bool):
            Include a 2.725 K contribution from the CMB (default True).
        """

        T_cmb = 2.725 if include_cmb else 0

        if freqs is not None:
            self.generate(freqs)

        pix = hp.ang2pix(
            self.nside, coords.galactic.l.deg, coords.galactic.b.deg, lonlat=True
        )
        if self.generated_map_data.ndim == 2:
            return self.generated_map_data[:, pix] + T_cmb
        else:
            return self.generated_map_data[pix] + T_cmb

    def write_fits(self, filename):
        """Write out map data as FITS file.

        Parameters
        ----------
        filename: str
            file name for output FITS file
        """
        hp.write_map(filename, self.generated_map_data, column_units=self.data_unit)



from .gsm08 import GlobalSkyModel, GSMObserver
from .gsm16 import GlobalSkyModel16, GSMObserver16
from .haslam import HaslamObserver, HaslamSkyModel
from .lfsm import LFSMObserver, LowFrequencySkyModel
from .component_data import download_map_data as download_map_data

# Add aliases
GlobalSkyModel08 = GlobalSkyModel
GSMObserver08 = GSMObserver

def init_gsm(gsm_name: str = "gsm08", *args, **kwargs):
    """Initialize a GDSM object by ID/name

    Returns a diffuse sky model (subclass of BaseSkyModel), based on one of:
      * **GSM08:** A model of diffuse Galactic radio emission from 10 MHz to 100 GHz,
                   [Oliveira-Costa et. al., (2008)](https://ui.adsabs.harvard.edu/abs/2008MNRAS.388..247D/abstract).
      * **GSM16:** An improved model of diffuse galactic radio emission from 10 MHz to 5 THz,
                   [Zheng et. al., (2016)](https://ui.adsabs.harvard.edu/abs/2017MNRAS.464.3486Z/abstract).
      * **LFSM:** The LWA1 Low Frequency Sky Survey (10-408 MHz)
                  [Dowell et. al. (2017)](https://ui.adsabs.harvard.edu/abs/2017MNRAS.469.4537D/abstract).
      * **Haslam:** A frequency-scaled model using a spectral index, based on the Haslam 408 MHz map
                   [Haslam 408 MHz](https://lambda.gsfc.nasa.gov/product/foreground/fg_2014_haslam_408_info.cfm).

    Args:
        gsm_name (str): One of 'gsm08', 'gsm16', 'lfsm' or 'haslam'

    Returns:
        sky_model (various): Corresponding sky model
    """
    gsm_name = gsm_name.lower().strip()

    if gsm_name == 'gsm': # Shorthand for GSM08
        return GlobalSkyModel(*args, **kwargs)
    elif gsm_name == 'gsm08':
        return GlobalSkyModel(*args, **kwargs)
    elif gsm_name == 'gsm16':
        return GlobalSkyModel16(*args, **kwargs)
    elif gsm_name == 'lfsm':
        return LowFrequencySkyModel(*args, **kwargs)
    elif gsm_name == 'haslam':
        return HaslamSkyModel(*args, **kwargs)
    else:
        raise ValueError(f'Invalid model specification "{gsm_name}"')



def init_observer(gsm_name: str = "gsm08", *args, **kwargs):
    """Initialize a GDSM Observer object by ID/name

    Returns an observer (subclass of BaseObserver), where the diffuse sky is created from one of:
      * **GSM08:** A model of diffuse Galactic radio emission from 10 MHz to 100 GHz,
                   [Oliveira-Costa et. al., (2008)](https://ui.adsabs.harvard.edu/abs/2008MNRAS.388..247D/abstract).
      * **GSM16:** An improved model of diffuse galactic radio emission from 10 MHz to 5 THz,
                   [Zheng et. al., (2016)](https://ui.adsabs.harvard.edu/abs/2017MNRAS.464.3486Z/abstract).
      * **LFSM:** The LWA1 Low Frequency Sky Survey (10-408 MHz)
                  [Dowell et. al. (2017)](https://ui.adsabs.harvard.edu/abs/2017MNRAS.469.4537D/abstract).
      * **Haslam:** A frequency-scaled model using a spectral index, based on the Haslam 408 MHz map
                   [Haslam 408 MHz](https://lambda.gsfc.nasa.gov/product/foreground/fg_2014_haslam_408_info.cfm).

    Args:
        gsm_name (str): One of 'gsm08', 'gsm16', 'lfsm' or 'haslam'

    Returns:
        observer (various): Corresponding sky model observer
    """
    gsm_name = gsm_name.lower().strip()

    if gsm_name == 'gsm': # Shorthand for GSM08
        return GSMObserver(*args, **kwargs)
    elif gsm_name == 'gsm08':
        return GSMObserver(*args, **kwargs)
    elif gsm_name == 'gsm16':
        return GSMObserver16(*args, **kwargs)
    elif gsm_name == 'lfsm':
        return LFSMObserver(*args, **kwargs)
    elif gsm_name == 'haslam':
        return HaslamObserver(*args, **kwargs)
    else:
        raise ValueError(f'Invalid model specification "{gsm_name}"')



from __future__ import annotations
from astropy.utils.data import conf, download_file
from astropy.config import get_cache_dir

# Zenodo can be slow to respond, so set timeout to 30 seconds
conf.remote_timeout = 30

FILE_HOST = "datacentral.org.au"
GSM_DATA_URL          = [
    "https://apps.datacentral.org.au/pygdsm/data/gsm_components.h5",
    "https://zenodo.org/record/3835582/files/gsm_components.h5?download=1"
    ]

GSM2016_DATA_URL      = [
    "https://apps.datacentral.org.au/pygdsm/data/gsm2016_components.h5",
    "https://zenodo.org/record/3835582/files/gsm2016_components.h5?download=1"
]

GSM2008_TEST_DATA_URL = [
    "https://apps.datacentral.org.au/pygdsm/data/gsm_fortran_test_data.h5",
    "https://zenodo.org/record/3835582/files/gsm_fortran_test_data.h5?download=1"
]

LFSM_DATA_URL         = [
    "https://apps.datacentral.org.au/pygdsm/data/lfsm.h5",
    "https://zenodo.org/record/3835582/files/lfsm.h5?download=1"
]

HASLAM_DATA_URL       = [
    "https://lambda.gsfc.nasa.gov/data/foregrounds/haslam_2014/haslam408_dsds_Remazeilles2014.fits"
]

DATA_URLS = [
            GSM_DATA_URL,
            GSM2016_DATA_URL,
            GSM2008_TEST_DATA_URL,
            LFSM_DATA_URL,
            HASLAM_DATA_URL
            ]

def download_from_url_list(url_list: list[str] | str):
    """Download from a list of URLs, trying each in turn.

    Args:
        url_list (list[str] | str): List of URLs to try, or a single URL.

    Returns:
        str: Path to the downloaded file.
    """
    if isinstance(url_list, str):
        url_list = [url_list]

    for url in url_list:
        try:
            return download_file(url, cache=True, show_progress=True)
        except Exception as e:
            print(f"Failed to download {url}: {e}")
    raise RuntimeError("All URLs failed to download.")


def download_map_data():
    """Download component data."""
    print("Checking for component data files...")
    for URL in DATA_URLS:
        download_from_url_list(URL)

    CACHE_PATH = get_cache_dir()
    print(f"Data saved in {CACHE_PATH}")


import sys

import numpy as np

sys.path.append("/workdata/pygdsm")

from datetime import datetime

import healpy as hp
import pylab as plt

from pygdsm import GSMObserver, HaslamObserver, HaslamSkyModel


def test_compare_gsm_to_old():
    gl = HaslamSkyModel(freq_unit="MHz")
    dl = gl.generate(408)
    gl.view()

    import pylab as plt

    plt.show()


def test_observer_test():
    # Setup observatory location - in this case, Parkes Australia
    (latitude, longitude, elevation) = ("-32.998370", "148.263659", 100)
    ov = HaslamObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    ov.generate(200)
    d = ov.view(logged=True)

    ov = GSMObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    ov.generate(200)
    d = ov.view(logged=True)
    plt.show()

    ov = GSMObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    horizon_elevation = 85.0
    ov.generate(1400, horizon_elevation=str(horizon_elevation))
    d_85deg_horizon = ov.view(logged=True)

    ov = GSMObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    ov.generate(1400, horizon_elevation=np.deg2rad(horizon_elevation))
    d_85deg2rad_horizon = ov.view(logged=True)

    assert np.all(d_85deg_horizon == d_85deg2rad_horizon), "The two methods for calculating the artificial horizon do not match."
    assert np.ma.count_masked(d_85deg_horizon).sum() == 3139749
    plt.show()

def test_cmb_removal():
    g = HaslamSkyModel(freq_unit="MHz", include_cmb=False)
    sky_no_cmb = g.generate(400)
    g = HaslamSkyModel(freq_unit="MHz", include_cmb=True)
    sky_with_cmb = g.generate(400)

    T_cmb = (sky_with_cmb - sky_no_cmb).mean()
    print(T_cmb)
    assert np.isclose(T_cmb, 2.725)


if __name__ == "__main__":
    test_compare_gsm_to_old()
    test_observer_test()



import healpy as hp
import numpy as np
from astropy import units
from astropy.utils.data import download_file

from .base_observer import BaseObserver
from .base_skymodel import BaseSkyModel
from .component_data import HASLAM_DATA_URL, download_from_url_list

T_CMB = 2.725


class HaslamSkyModel(BaseSkyModel):
    """Haslam destriped, desourced sky model"""

    def __init__(self, freq_unit="MHz", spectral_index=-2.6, include_cmb=False):
        """Generate a sky model at a given frequency based off Haslam 408 MHz map

        Parameters
        ----------
        freq_unit (str): Frequency unit to use, defaults to MHz
        spectral_index (float): Spectral index to use for calculations.
        include_cmb (bool):  Choose whether to include the CMB. Defaults to False. A value of
                             T_CMB = 2.725 K is used if True.
        Notes
        -----
        The basemap is the destriped and desoruced Haslam map from Remazeilles (2015)
        A T_CMB value of 2.725 K is subtracted to account for the microwave background component.

        This is a crude model; the sky's spectral index changes with sidereal time
        and a singular spectral index is not realistic. Here are some measured values
        at different frequencies:

            Frequency        Spectral Index        Reference     Notes
            50--100 MHz     si = -2.56 +/- 0.03    Mozdzen+19    LST 0--12 hr
                                  2.46 +/- 0.01                  LST 18.2 hr
            90--190 MHz     si = -2.61 +/- 0.01    Mozdzen+16    LST 0--12 hr
                                 −2.50 +/- 0.02                  LST 17.7 hr
            0.4--4.7 GHz 	si = −2.85 +/- 0.05    Dickinson+19
            0.4--22.8 GHz 	si = −2.88 +/- 0.03    Dickinson+19
            4.7--22.8 GHz 	si = −2.91 +/- 0.04    Dickinson+19
            22.8--44 GHz 	si = −2.85 +/- 0.14    Dickinson+19

        References
        ----------
        Remazeilles et al (2015), improved Haslam map, https://doi.org/10.1093/mnras/stv1274
        Mozdzen et al (2016), EDGES high-band, https://doi.org/10.1093/mnras/stw2696
        Mozdzen et al (2019), EDGES low-band, https://doi.org/10.1093/mnras/sty3410
        Dickinson et al (2019), C-BASS experiment, https://doi.org/10.1093/mnras/stz522
        """
        data_unit = "K"
        basemap = "Haslam"

        # download component data as needed using astropy cache
        HASLAM_FILEPATH = download_from_url_list(HASLAM_DATA_URL)

        super(HaslamSkyModel, self).__init__(
            "Haslam", HASLAM_FILEPATH, freq_unit, data_unit, basemap
        )
        self.spectral_index = spectral_index
        self.data = hp.read_map(self.fits, dtype=np.float64) - T_CMB
        self.nside = 512

        self.include_cmb = include_cmb

    def generate(self, freqs):
        """Generate a global sky model at a given frequency or frequencies

        Parameters
        ----------
        freqs: float or np.array
            Frequency for which to return GSM model

        Returns
        -------
        gsm: np.array
            Global sky model in healpix format, with NSIDE=256. Output map
            is in galactic coordinates, and in antenna temperature units (K).

        """
        # convert frequency values into Hz
        freqs = np.array(freqs) * units.Unit(self.freq_unit)
        freqs_mhz = freqs.to("MHz").value

        if isinstance(freqs_mhz, float):
            freqs_mhz = np.array([freqs_mhz])

        map_out = np.outer(
            (freqs_mhz / 408.0) ** (self.spectral_index), self.data
        ).squeeze()

        if self.include_cmb:
            map_out += T_CMB
        self.generated_map_data = map_out
        self.generated_map_freqs = freqs
        return map_out


class HaslamObserver(BaseObserver):
    def __init__(self):
        """Initialize the Observer object.

        Calls ephem.Observer.__init__ function and adds on gsm
        """
        super(HaslamObserver, self).__init__(gsm=HaslamSkyModel)



"""
test_base_skymodel.py
=====================

Tests for GSM base skymodel
"""

from pathlib import Path

import pytest

from pygdsm.base_skymodel import BaseSkyModel
from pygdsm.component_data import GSM_DATA_URL, download_from_url_list

# download component data as needed using astropy cache
GSM_FILEPATH = download_from_url_list(GSM_DATA_URL)

def test_base_skymodel_init():
    gsm = BaseSkyModel(
        "TEST_GSM", GSM_FILEPATH, freq_unit="MHz", basemap="haslam", data_unit="K"
    )

    with pytest.raises(RuntimeError):
        gsm = BaseSkyModel(
            "TEST_GSM",
            "not_a_file.madeup",
            freq_unit="MHz",
            basemap="haslam",
            data_unit="K",
        )

    with pytest.raises(RuntimeError):
        gsm = BaseSkyModel(
            "TEST_GSM",
            Path(__file__).absolute(),
            freq_unit="MHz",
            basemap="haslam",
            data_unit="K",
        )


if __name__ == "__main__":
    test_base_skymodel_init()



import sys

sys.path.append("/workdata/pygdsm")

from datetime import datetime

import healpy as hp
import numpy as np
import pylab as plt
import pytest

from pygdsm import GlobalSkyModel, GlobalSkyModel16, GSMObserver, GSMObserver16


def test_compare_gsm_to_old():
    g = GlobalSkyModel16(freq_unit="GHz", resolution="hi", data_unit="TCMB")
    d = g.generate(0.408)
    g.view()

    g_old = GlobalSkyModel(freq_unit="GHz", basemap="haslam")
    d_old = g_old.generate(0.408)
    g_old.view()


def test_observer_test():
    # Setup observatory location - in this case, Parkes Australia
    (latitude, longitude, elevation) = ("-32.998370", "148.263659", 100)
    ov = GSMObserver16()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    ov.generate(1400)
    d = ov.view(logged=True)

    ov = GSMObserver()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    ov.generate(1400)
    d = ov.view(logged=True)
    plt.show()

    ov = GSMObserver16()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    horizon_elevation = 85.0
    ov.generate(1400, horizon_elevation=str(horizon_elevation))
    d_85deg_horizon = ov.view(logged=True)

    ov = GSMObserver16()
    ov.lon = longitude
    ov.lat = latitude
    ov.elev = elevation
    ov.date = datetime(2000, 1, 1, 23, 0)

    ov.generate(1400, horizon_elevation=np.deg2rad(horizon_elevation))
    d_85deg2rad_horizon = ov.view(logged=True)

    assert np.all(d_85deg_horizon == d_85deg2rad_horizon), "The two methods for calculating the artificial horizon do not match."
    assert np.ma.count_masked(d_85deg_horizon).sum() == 12558953
    plt.show()

def test_interp():
    f = np.arange(40, 80, 5)
    for interp in ("pchip", "cubic"):
        for SkyModel in (GlobalSkyModel, GlobalSkyModel16):
            name = str(SkyModel).strip("<>").split(".")[-1].strip("' ")
            gsm = SkyModel(freq_unit="MHz", interpolation=interp)
            d = gsm.generate(f)

            sky_spec = d.mean(axis=1)
            fit = np.poly1d(np.polyfit(f, sky_spec, 5))(f)

            plt.plot(f, sky_spec - fit, label=f"{name}: {interp}")

    plt.xlabel("Frequency [MHz]")
    plt.ylabel("Residual [K]")
    plt.legend()
    plt.show()


def test_gsm_opts():
    g = GlobalSkyModel16(freq_unit="GHz", resolution="hi", data_unit="TCMB")
    d = g.generate(0.408)
    g = GlobalSkyModel16(freq_unit="GHz", resolution="lo", data_unit="MJysr")
    d = g.generate(0.408)
    g = GlobalSkyModel16(freq_unit="MHz", resolution="lo", data_unit="TRJ")
    d = g.generate(408)

    with pytest.raises(RuntimeError):
        g.generate(5e12)

    with pytest.raises(RuntimeError):
        g = GlobalSkyModel16(resolution="oh_hai")

    with pytest.raises(RuntimeError):
        g = GlobalSkyModel16(data_unit="furlongs/fortnight")


def test_cmb_removal():
    g = GlobalSkyModel16(
        freq_unit="GHz", resolution="lo", data_unit="TCMB", include_cmb=False
    )
    sky_no_cmb = g.generate(400)
    g = GlobalSkyModel16(
        freq_unit="GHz", resolution="lo", data_unit="TCMB", include_cmb=True
    )
    sky_with_cmb = g.generate(400)

    T_cmb = (sky_with_cmb - sky_no_cmb).mean()
    print(T_cmb)
    assert np.isclose(T_cmb, 2.725)


if __name__ == "__main__":
    test_compare_gsm_to_old()
    test_observer_test()
    test_gsm_opts()
    test_interp()



"""A setuptools based setup module."""

# Always prefer setuptools over distutils
# To use a consistent encoding
from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.read().splitlines()

with open("requirements_test.txt", "r") as fh:
    test_requirements = fh.read().splitlines()

setup(
    name="pygdsm",
    version="1.6.4",
    description="Python Global Sky Model of diffuse Galactic radio emission",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://github.com/telegraphic/pygdsm",
    # Author details
    author="Danny C. Price",
    author_email="daniel.price@skao.int",
    # Choose your license
    license="MIT",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    # What does your project relate to?
    keywords="radio astronomy sky model galactic diffuse emission",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,
    tests_require=test_requirements,
)
